#!/usr/bin/perl
#
# $Id: macblock.pl,v 1.6 2009-06-03 13:35:55 obiwan Exp $
#
# Script to fetch allowed IPs from a mysql db
# parse the output from brconfig $bn and compare them
# then enter the rogue macs into the database and possibly
# create a new bridgename.$bn and reread it.
#
# (c) 2003-2009 SCS TeleMedia AG
# maintained cschnee@telemedia.ch
#
# TODO:
# -a bit more of help...
#
# compile with
# perl2exe  -I=/usr/libdata/perl5:/usr/local/libdata/perl5/site_perl/i386-openbsd:/usr/local/libdata/perl5/site_perl macblock.pl


##############################################
# Modules to use..
##############################################
use DBD::mysql;
use Getopt::Std;
use Config::IniFiles;

# explicit mention for p2x
use attributes;
use Carp::Heavy;
##############################################
# check command line switches
##############################################
foreach (@ARGV) {
	SWITCH: {
		if ($_ eq "-?" || $_ eq "-h") {$usage_info = "1"; $verbose = "1"; last SWITCH; }
		if ($_ eq "-vv")              {$dbg = "1"; last SWITCH; }
		if ($_ eq "-v") 		          {$verbose = "1"; last SWITCH; }
		if ($_ eq "-V")        		    {$vers = "1"; last SWITCH; }
	}
}

##############################################
# find version from cvs tag
##############################################
$version = '$Revision: 1.6 $';
$version =~ s/(Revision: )|(\$)|( \$$)//g;

##############################################
# display help text
##############################################
my $helptxt = <<"EOH";
macblock $version - (c) 2003-2009 SCS TeleMedia AG

This is the help for macblock

-v  be verbose
-vv debugging (much more verbose)
-V  print version numb of macblock

EOH

if ($usage_info eq "1") {
  display("$helptxt", v);
  exit(0);
}
if ( $vers eq "1" ) {
  print "macblock $version - (c) 2003-2004 SCS TeleMedia AG\n";
  exit(0);
}

########################################
# find out how and where we are
########################################
my $own_name = $0;
my $inifile = $own_name;
$inifile =~ s/\.[a-z0-9]+$//i;
if ( $inifile =~ /^\./) {
  $inifile =~ s/^\.//;
  my $cur_path = `pwd`;
  chomp ($cur_path);
  $inifile = $cur_path . $inifile;
  
}
$inifile =~ s/\/\///g;
$inifile .= '.ini';
if (!-e $inifile) {
	$inifile = '/etc/macblock.ini'
}
# remove path from exe/pl
$own_name =~ s/(\/[a-z0-9]+)+\///ig;
# remove suffix from exe/pl
$own_name =~ s/\.[a-z0-9]+$//ig;
display("Running as $own_name with ini $inifile", d);
# check for inifile existence
if (!-e $inifile) {
	die "$inifile not found!\n";
}

########################################
# read in ini file
########################################
my $ini = Config::IniFiles->new( -file => "$inifile" ) or die ("Can't open inifile $inifile: $!\n");
# Main
my $hot_if = $ini->val('Main', 'hot_if');
my $cool_if = $ini->val('Main', 'cool_if');
my $bridge_name = $ini->val('Main', 'bridgename');
my $btype =  $ini->val('Main', 'btype');
my $bid =  $ini->val('Main', 'bid');
# DB
my $db_srv = $ini->val( 'DB', 'host' );
my $db_user = $ini->val( 'DB', 'user' );
my $db_pass = $ini->val( 'DB', 'pass' );
my $db_name = $ini->val( 'DB', 'database' );
my $db_driver = $ini->val( 'DB', 'driver' );
my $db_allow_tbl = $ini->val( 'DB', 'allow_tbl' );
my $db_seen_tbl = $ini->val( 'DB', 'seen_tbl' );
my $db_dsn = "DBI:$db_driver:database=$db_name;host=$db_srv";
my $maxage = $ini->val( 'Bridge', 'maxage' );
my $fwddelay = $ini->val( 'Bridge', 'fwddelay' );
my $timeout = $ini->val( 'Bridge', 'timeout' );
my $maxaddr = $ini->val( 'Bridge', 'maxaddr' );
my $mpriority = $ini->val( 'Bridge', 'mpriority' );
my $spriority = $ini->val( 'Bridge', 'spriority' );
if ( $btype == 1 ) { $bprio = $mpriority }
else { $bprio = $spriority }



########################################
# bridge base and tail rules
########################################
my $brcfg_base = <<"EOBR";
########################################
# Automatically created by macblock.pl #
# Do not modify                        #
########################################
down
#flushall
del $cool_if
del $hot_if
add $cool_if
add $hot_if
#stp $cool_if
#stp $hot_if
#-discover $hot_if
maxage $maxage
fwddelay $fwddelay
timeout $timeout
maxaddr $maxaddr
priority $bprio
rule pass in  on $cool_if
rule pass out on $cool_if
# allow discovery / arp req
rule pass out on $hot_if dst ff:ff:ff:ff:ff:ff
EOBR

my $brcfg_tail = <<"EOTAIL";
rule pass in  on $hot_if tag unknown
rule pass out on $hot_if tag unknown
up
#EOF
EOTAIL

my $tmp_config_file = '/tmp/.brconfig_new';
my $real_config_file = "/etc/bridgename.$bridge_name";


##############################################
# Main Script
##############################################

##############################################
# fill some needed arrays and hashes
##############################################
my %br_mac = get_brc_mac($bridge_name, $hot_if);
my @br_mac = keys(%br_mac);
my %db_allow_mac = get_db_mac($db_dsn, $db_allow_tbl);
my @db_allow_mac = keys(%db_allow_mac);
my %db_seen_mac = get_db_mac($db_dsn, $db_seen_tbl);
my @db_seen_mac = keys(%db_seen_mac);

##############################################
# check which hosts seen on bridge are not
# in the seen table yet.
##############################################
@newmac = check_new_mac(\@db_seen_mac, \@br_mac);

# insert new macs into the seen table
foreach $item (@newmac) {
  display ("NEW MAC!!! : $item", v);
}
ins_db_mac(\@newmac,$db_seen_tbl);

# check if updates are desired
if (checkup()) {
  my @mac_allow = make_config();
  # write config file
  write_config(@mac_allow);
  # activate config
  do_activate();
  # reset update flag
  reset_upd_flag();
}



########################################
# Subs
########################################
##############################################
# Parse brconfig
##############################################
sub get_brc_mac {
  my $bridge = $_[0];
  my $ifname = $_[1];
  my %mac;
  my @brconfig_out = `brconfig $bridge`;
  foreach my $line (@brconfig_out) {
    chop($line);
    my @ldata = split(/ /, $line);
    if ($line =~ /^[\t]+[A-Ea-e0-9]{2,2}:.+:/) {
      $ldata[0] =~ s/[\t]//gi;
      if ($ldata[1] eq $ifname) {
        $mac{$ldata[0]} = $ldata[1];
        display ("$^T is a line with desired MAC: $ldata[0] ->> $ldata[1]", d);
      }
      else {
        display ("$^T is a line with a undesired MAC: $ldata[0] ->> $ldata[1]", d);
      }
    }
  }

  return %mac;
}

##############################################
# Get Macs from DB
##############################################
sub get_db_mac {
  my $q_handle;
  my %mac;
  my $db_handle = DBI->connect($db_dsn, $db_user, $db_pass) or die ("Can't connect to database $db_name on $db_srv :$!\n");
  my $querystring = "SELECT * FROM $_[1] WHERE 0=0;";
  $q_handle = $db_handle->prepare($querystring) or die ("Failed preparing SQL: $q_handle->errstr");
  $q_handle->execute or display ("Failed executing: " .$q_handle->errstr, 'v');
  while (my $ref = $q_handle->fetchrow_arrayref) {
    $mac{$$ref[mac]} = $hot_if;
  }
  $q_handle->finish();
  return %mac;
}

##############################################
# Insert Macs into a $table
##############################################
sub ins_db_mac {
  my ($newmac_ref, $table) = @_;
  my $rightnow = time();
  my $q_handle;
  my %mac;
  my $db_handle = DBI->connect($db_dsn, $db_user, $db_pass) or die ("Can't connect to database $db_name on $db_srv :$!\n");
  foreach my $newmac (@$newmac_ref) {
    my $querystring = "INSERT INTO $_[1] (mac, tstamp) VALUES ('$newmac', $rightnow);";
    $q_handle = $db_handle->prepare($querystring) or die ("Failed preparing SQL: $q_handle->errstr");
    $q_handle->execute or display ("Failed executing: " .$q_handle->errstr, 'v');
  }
  if ($q_handle) {
    $q_handle->finish();
  }
  return true;
}

##############################################
# Check for updates..
##############################################
sub checkup{
  my $db_handle = DBI->connect($db_dsn, $db_user, $db_pass) or die "Can't connect to database $db_name on $db_srv :$!\n";
  my $querystring = "SELECT do_update$bid FROM config WHERE 0=0;";
  $q_handle = $db_handle->prepare($querystring) or die ("Failed preparing SQL: $q_handle->errstr");
  $q_handle->execute or display ("Failed executing: " .$q_handle->errstr, 'v');
  my $ref = $q_handle->fetchrow_arrayref or die ("Failed to fetch row: $q_handle->errstr");
  display ("UPDATE = " . $$ref["do_update$bid"], d);
  if ($$ref["do_update$bid"] > 0) {
    display ("Update -> 1, activating changes...", v);
    $q_handle->finish();
    return 1;
  }
  else {
    display ("Update -> 0, sleeping...", v);
    $q_handle->finish();
    return 0;
  }
}

##############################################
# compare one array of macs against the other
# whereas the question is which of $_[0] are not in $_[1]
# and those are returned
##############################################
sub check_new_mac {
  my ($k_mac_ref, $n_mac_ref) = @_;
  my %seen;
  my @newmac;
  foreach my $kmac (@$k_mac_ref) {
    display ("SEEN MAC: $kmac", d);
    $seen{$kmac} = 1;
  }
  foreach my $nmac (@$n_mac_ref) {
    unless ($seen{$nmac}) {
      display ("FOUND NEW MAC: $nmac", v);
      push(@newmac,$nmac);
    }
  }
  return @newmac;
}

##############################################
# create config
##############################################
sub make_config {
  my $now = localtime();
  my @mac_cfg = ("#machine generated config from $now");
  my $brline_in;
  my $brline_out;
  my $db_handle = DBI->connect($db_dsn, $db_user, $db_pass) or die ("Can't connect to database $db_name on $db_srv :$!\n");
  my $querystring = "SELECT mac FROM mac_addr_allow WHERE 0=0;";
  $q_handle = $db_handle->prepare($querystring) or die ("Failed preparing SQL: $q_handle->errstr");
  $q_handle->execute or display ("Failed executing: " .$q_handle->errstr, 'v');
  while (my $ref = $q_handle->fetchrow_arrayref) {
    $brline_in = "rule pass in on $hot_if src $$ref[mac] tag allowed";
    $brline_out = "rule pass out on $hot_if dst $$ref[mac] tag allowed";
    push(@mac_cfg, $brline_in, $brline_out);
  }
  $q_handle->finish();
  push(@mac_cfg, '#end of generated cfg');
  return @mac_cfg;
}

##############################################
# write config
##############################################
sub write_config {
  my @brcfg = @_;
  # remove old tmp file first...
  unlink($tmp_config_file);
  open(NEWBRCFG, "> $tmp_config_file");
  # add base rules...
  print NEWBRCFG $brcfg_base;
  foreach my $cfgline (@brcfg) {
    print NEWBRCFG "$cfgline\n";
    display ("CFG: $cfgline", d);
  }
  # add tail rules...
  print NEWBRCFG $brcfg_tail;
  close(NEWBRCFG);
}

##############################################
# activate brconfig
##############################################
sub do_activate {
  # move da file in place
  # mount / rw
  # rename($tmp_config_file, $real_config_file);
  # mount / ro
  #$result = system("sudo mount -o rw /");
  #if ($result > 0) {
  #  display ("Cannot mount FS rw : $result", v);
  #  $result = system("sudo mount -o ro /");
  #  exit(255);
  #}
  my $result = system("sudo mv -f $tmp_config_file $real_config_file");
  if ($result > 0) {
    display ("Update failed with : $result", v);
    #$result = system("sudo mount -o ro /");
    exit(255);
  }
  # change perms on real config
  $result = system("sudo chmod 644 $real_config_file");
  if ($result > 0) {
    display ("Update failed with : $result", v);
    #$result = system("sudo mount -o ro /");
    exit(255);
  }
  #$result = system("sudo mount -o ro /");
  # call 'sh /etc/netstart bridge0'
  $result = system("sudo sh /etc/netstart $bridge_name");
  if ($result > 0) {
    display ("Update failed with : $result", v);
    exit(255);
  }

}

##############################################
# reset update flag
##############################################
sub reset_upd_flag {
  my $db_handle = DBI->connect($db_dsn, $db_user, $db_pass) or die ("Can't connect to database $db_name on $db_srv :$!\n");
  my $querystring = "UPDATE config SET do_update$bid=0 WHERE 0=0;";
  $q_handle = $db_handle->prepare($querystring) or die ("Failed preparing SQL: $q_handle->errstr");
  $q_handle->execute or display ("Failed executing: " .$q_handle->errstr, 'v');
  $q_handle->finish();
}

##############################################
# debug display
##############################################
sub display {
	if ($verbose) {
		if ($_[2] == 2 && $_[1] eq 'v') {
			print "$_[0] ";
		}
		elsif ($_[1] eq 'v') {
			print "$_[0] \n";
		}
	}
	if ($dbg) {
		if ($_[1] == 1) {
			print "$_[0] ";
		}
		else {
			print "$_[0] \n";
		}
	}
}

